from math import ceil, pi
import pandas as pd


def tower_mass(nameplate, HH):
    # nameplate = nameplate_MW * 1000
    tower_mass_tonne = (
                        nameplate / ((-0.0408 * (nameplate ** 2)) +
                                     (0.6734 * nameplate) + 0.632)
                       ) * \
                       (0.0401 * (HH ** 2)) - (4.4775 * HH) + 230.88
    return tower_mass_tonne


# number of tower sections:
def tower_specs(HH, tower_mass):
    number_of_sections = max(ceil(HH/30), ceil(tower_mass/80))
    # print(number_of_sections)
    tower_section_height = HH/number_of_sections
    return number_of_sections, tower_section_height


# TODO: Add additional argument (hub height) for storing component's lift
#  height and lever arm m.
def edit_tower_sections(components_DF,
                        number_of_sections,
                        tower_mass_tonne,
                        tower_section_height
                        ):
    tower_section_mass = tower_mass_tonne / number_of_sections

    remove_existing_tower_sections = components_DF[components_DF[
                                        'Component'].str.match('Tower')].index
    components_DF = components_DF.drop(components_DF.index[
                                            remove_existing_tower_sections])

    lever_arm_m = list()
    i = 0
    while i < number_of_sections:
        lift_height_m = (i+1) * tower_section_height
        tower_radius = 1.35  # in m.
        # TODO: confirm value of tower radius. I've read this is closer to 4.5m
        surface_area_m2 = pi * tower_section_height * (tower_radius**2)
        if i == 0:
            lever_arm_m.append(0.5 * tower_section_height)
        else:
            lever_arm_m.append(
                                (i * tower_section_height) +
                                (0.5 * tower_section_height)
                               )

        tower_section_df = pd.DataFrame(
            [['Tower section ' + str(i + 1),
              tower_section_mass,
              lift_height_m,
              surface_area_m2,
              0.6,
              1.1,
              tower_section_height,
              lever_arm_m[i],
              1,
              6,
              0.5,
              0,
              1]],
            columns=['Component',
                     'Mass tonne',
                     'Lift height m',
                     'Surface area sq m',
                     'Coeff drag',
                     'Coeff drag (installed)',
                     'Section height m',
                     'Lever arm m',
                     'Cycle time installation hrs',
                     'Offload hook height m',
                     'Offload cycle time hrs',
                     'Multplier drag rotor',
                     'Multiplier tower drag'])

        components_DF = components_DF.append(tower_section_df)
        i += 1
    return components_DF


def blade_mass_ton(rotor_diameter_m):
    combined_blade_mass_ton = (0.0056 * (rotor_diameter_m ** 2)) - \
                              (0.3818 * rotor_diameter_m) + \
                              11.753
    return combined_blade_mass_ton


def edit_blade_info(components_DF, blade_mass_ton, HH, rotor_diameter_m):
    # https://www.slideshare.net/pawanrm1/wind-turbine-design
    rotor_solidity = 0.1
    swept_area_m2 = pi * (rotor_diameter_m ** 2)
    blade_surface_area_m2 = rotor_solidity * swept_area_m2 / 3
    single_blade_mass_ton = blade_mass_ton / 3  # since LandBOSSE assumes
    # it's a 3-bladed machine
    remove_existing_blade_rows = components_DF[
                        components_DF['Component'].str.match('Blade')].index

    components_DF = components_DF.drop(
                            components_DF.index[remove_existing_blade_rows])

    lift_height_m = HH
    lever_arm_m = HH

    blade_DF = pd.DataFrame(
        [['Blade 1',
          single_blade_mass_ton,
          lift_height_m,
          blade_surface_area_m2,
          0.1,
          1.4,
          0,
          lever_arm_m,
          1,
          6,
          0.5,
          0.667,
          0]],
        columns=[
            'Component',
            'Mass tonne',
            'Lift height m',
            'Surface area sq m',
            'Coeff drag',
            'Coeff drag (installed)',
            'Section height m',
            'Lever arm m',
            'Cycle time installation hrs',
            'Offload hook height m',
            'Offload cycle time hrs',
            'Multplier drag rotor',
            'Multiplier tower drag'])

    components_DF = components_DF.append(blade_DF)  # blade 1
    components_DF = components_DF.append(blade_DF)  # blade 2
    components_DF = components_DF.append(blade_DF)  # blade 3
    return components_DF


def hub_mass(nameplate):
    nameplate = nameplate * 1000    # convert form MW to kW
    hub_mass_ton = ((1e-06) * (nameplate ** 2)) + (0.0039 * nameplate) + 3.652
    return hub_mass_ton


def edit_hub_info(components_DF, hub_mass, HH):
    remove_existing_hub_rows = components_DF[components_DF[
                                        'Component'].str.match('Hub')].index
    components_DF = components_DF.drop(
                                components_DF.index[remove_existing_hub_rows])
    lift_height_m = HH

    # TODO: find a relationship for hub surface area.
    hub_surface_area = 12.16
    lever_arm_m = HH
    hub = pd.DataFrame(
        [['Hub',
          hub_mass,
          lift_height_m,
          hub_surface_area,
          1.1,
          1.1,
          0,
          lever_arm_m,
          1,
          6,
          0.5,
          0,
          0]],
        columns=[
            'Component',
            'Mass tonne',
            'Lift height m',
            'Surface area sq m',
            'Coeff drag',
            'Coeff drag (installed)',
            'Section height m',
            'Lever arm m',
            'Cycle time installation hrs',
            'Offload hook height m',
            'Offload cycle time hrs',
            'Multplier drag rotor',
            'Multiplier tower drag'])
    components_DF = components_DF.append(hub)
    return components_DF


def nacelle_mass(nameplate):
    nameplate = nameplate * 1000  # convert form MW to kW
    nacelle_mass_ton = (4e-06 * (nameplate**2)) + (0.0233 * nameplate) + 3.7349
    return nacelle_mass_ton


def edit_nacelle_info(components_DF, nacelle_mass_ton, HH):
    components_DF = components_DF[components_DF.Component != 'Nacelle']
    lift_height_m = HH

    # TODO: find a relationship for nacelle surface area.
    nacelle_surface_area = 33
    lever_arm_m = HH
    nacelle = pd.DataFrame(
        [['Nacelle',
          nacelle_mass_ton,
          lift_height_m,
          nacelle_surface_area,
          0.8,
          0.8,
          0,
          lever_arm_m,
          1.5,
          6,
          0.5,
          1,
          0]],
        columns=[
            'Component',
            'Mass tonne',
            'Lift height m',
            'Surface area sq m',
            'Coeff drag',
            'Coeff drag (installed)',
            'Section height m',
            'Lever arm m',
            'Cycle time installation hrs',
            'Offload hook height m',
            'Offload cycle time hrs',
            'Multplier drag rotor',
            'Multiplier tower drag'])

    components_DF = components_DF.append(nacelle)
    return components_DF



import os


def landbosse_test_input_dir():
    """
    This function is to find the input directory of the landbosse.

    This function checks to see if the LANDBOSSE_INPUT_DIR environment
    variable is defined. If so, it returns the value of that variable.

    If the environment variable is not defined, it returns 'inputs'
    relative to the current working directory.

    Returns
    -------
    str
        The input directory.
    """

    test_source_dir = os.path.dirname(os.path.realpath(__file__))
    default_test_inputs_dir = os.path.join(test_source_dir, 'inputs')
    return os.environ.get('LANDBOSSE_TEST_INPUT_DIR', default_test_inputs_dir)




import numpy as np
import pandas as pd

"""
This module contains the logic to handle a tree to compute
points in an N-dimensional parametric search space.
"""

class GridSearchTreeNode:
    """
    This just contains information about a node in the grid
    search tree.
    """

    def __init__(self):
        self.cell_specification = None
        self.children = []
        self.value = None


class GridSearchTree:
    """
    This class implements a k-ary tree to compute possible
    combinations of points in a N-dimensional parametric
    search space.
    """

    def __init__(self, parametric_list):
        """
        This simply sets the parametric_list. See the first dataframe
        described in the docstring of XlsxReader.create_parametric_value_list()

        Parameters
        ----------
        parametric_list : pandas.DataFrame
            The dataframe of the parametrics list.
        """
        self.parametric_list = parametric_list

    def build_grid_tree_and_return_grid(self):
        """
        See the dataframes in XlsxReader.create_parametric_value_list()
        for context.

        This builds a tree of points in the search space and traverse
        it to find points on the grid.

        Returns
        -------
        """

        # Build the tree. Its leaf nodes contain the values for each
        # point in the grid.
        root = self.build_tree()

        # Recursions of the traversal method needs to start with an empty
        # list.
        grid = self.dfs_search_tree(root, traversal=[])
        return grid

    def build_tree(self, depth=0, root=None):
        """
        This method builds a k-ary tree to contain cell_specifications and
        their values.

        Callers from outside this method shouldn't override the defaults
        for the parameters. These parameters are to manage the recursion,
        and are supplied by this method when it invokes itself.

        Parameters
        ----------
        root : GridSearchTreeNode
            The root of the subtree. At the start of iteration, at the
            root of the whole tree, this should be None.

        depth : int
            The level of the tree currently being built. This is
            also the row number in the dataframe from which the tree
            is being built.

        Returns
        -------
        GridSearchTreeNode
            The root of the tree just built.
        """
        row = self.parametric_list.iloc[depth]
        cell_specification = f"{row['Dataframe name']}/{row['Row name']}/{row['Column name']}"

        # First, make an iterable of the range we are going to be using.
        if 'Value list' in row and not pd.isnull(row['Value list']):
            values = [float(value) for value in row['Value list'].split(',')]
        else:
            start = row['Min']
            end = row['Max']
            step = row['Step']
            values = np.arange(start, end + step, step)

        if root == None:
            root = GridSearchTreeNode()

        # Putting the stop at end + step ensures the end value is in the sequence
        #
        # Append children for each value in the parametric step sequence.

        for value in values:
            child = GridSearchTreeNode()
            child.value = value
            child.cell_specification = cell_specification
            root.children.append(child)

            # If there are more levels of variables to add, recurse
            # down 1 level.
            if len(self.parametric_list) > depth + 1:
                self.build_tree(depth + 1, child)

        return root

    def dfs_search_tree(self, root, traversal, path=None):
        """
        This does a depth first search traversal of the GridSearchTree
        specified by the root parameter. It stores the node it encounters
        in the list referenced by traversal.

        There is a distinction from normal DFS traversals: Only leaf nodes
        are recorded in the traversal. This means that only nodes that have
        a complete list of cell specifications and values are returned.

        Parameters
        ----------
        root : GridSearchTreeNode
            The root of the

        traversal : list
            The nodes traversed on the tree. When this method is called
            by an external caller, this should be an empty list ([])

        path : list
            This shouldn't be manipulated except by this method itself.
            It is for storing the paths to the leaf nodes.

        Returns
        -------
        list
            A list of dictionaries that hold the cell specifications and
            values of each leaf node.
        """

        path = [] if path is None else path[:]

        if root.cell_specification is not None:
            path.append({
                'cell_specification': root.cell_specification,
                'value': root.value,
            })

        if len(root.children) == 0:
            traversal.append(path)

        for child in root.children:
            self.dfs_search_tree(child, traversal, path)

        return traversal



from .ManagementCost import ManagementCost
from .Manager import Manager
from .WeatherDelay import WeatherDelay
from .FoundationCost import FoundationCost
from .ErectionCost import ErectionCost
from .SitePreparationCost import SitePreparationCost
from .SubstationCost import SubstationCost
from .GridConnectionCost import GridConnectionCost
from .CollectionCost import Cable, Array, ArraySystem
from .DevelopmentCost import DevelopmentCost
from .DefaultMasterInputDict import DefaultMasterInputDict



from .XlsxReader import XlsxReader
from .XlsxGenerator import XlsxGenerator
from .XlsxManagerRunner import XlsxManagerRunner
from .XlsxSerialManagerRunner import XlsxSerialManagerRunner
from .XlsxParallelManagerRunner import XlsxParallelManagerRunner
from .XlsxFileOperations import XlsxFileOperations
from .XlsxValidator import XlsxValidator
from .XlsxDataframeCache import XlsxDataframeCache
from .CsvGenerator import CsvGenerator



from unittest import TestCase
import os

import numpy as np
import pandas as pd

from landbosse.model import FoundationCost
from landbosse.tests.model.test_WeatherDelay import generate_a_year
from landbosse.tests.model.test_filename_functions import landbosse_test_input_dir

pd.set_option('display.width', 6000)
pd.set_option('display.max_columns', 20)


class TestFoundtionCost(TestCase):
    def setUp(self):
        """
         This setUp() method executes before each test. It creates an
         instance attribute, self.instance, that refers to a FoundationCost
         instance under test. input_dict is the dictionary of inputs values
         that configures that instance.
         This is just a default dictionary. Methods that test behavior of
         specific variables customize the relevant key/value pairs.
         The key site_facility_building_area_df is set to None here. The
         dataframe is read from the csv in the site_facility tests below.
        """


        self.input_dict = dict()
        self.input_dict['depth'] = 3.05 # in m
        self.project_name = 'project_1'

        self.input_dict['rated_thrust_N'] = 742e3
        self.input_dict['bearing_pressure_n_m2'] = 244200
        self.input_dict['critical_velocity_m_per_s'] = 52.5
        self.input_dict['gust_velocity_m_per_s'] = 60

        
        #Below are the inputs for calculate_foundation_load():
        component_data = os.path.join(landbosse_test_input_dir(), 'components.csv')
        self.input_dict['component_data'] = pd.read_csv(component_data)
        crew_data = os.path.join(landbosse_test_input_dir(), 'crews.csv')
        self.input_dict['crew'] = pd.read_csv(crew_data)
        crew_cost_data = os.path.join(landbosse_test_input_dir(), 'crew_price_proprietary.csv')
        self.input_dict['crew_cost'] = pd.read_csv(crew_cost_data)
        for component in self.input_dict['component_data'].keys():
            self.input_dict[component] =  np.array(self.input_dict['component_data'][component])



        # Below are the inputs for determine_foundation_size():
        #'Radius_m' -> Calculated in 'calculate_foundation_load()'
        #'depth'

        #Below are the inputs for 'estimate_material_needs()':
        self.input_dict['num_turbines'] = 100
        #foundation_volume_m3_per_turbine -> calculated in 'determine_foundation_size()':

        #Below are the inputs for 'estimate_construction_time()':
        self.input_dict['duration_construction'] = 9    #months
        self.input_dict['construct_duration'] = 9 #months
        self.input_dict['hour_day'] = {'long': 24, 'normal': 10}

        #Weather window:
        self.input_dict['num_delays'] = 7
        self.input_dict['avg_hours_per_delay'] = 20
        self.input_dict['std_dev_hours_per_delay'] = 5
        self.input_dict['delay_speed_m_per_s'] = 9
        self.input_dict['seed'] = 101
        self.input_dict['weather_window'] = generate_a_year(num_delays=self.input_dict['num_delays'],
                                              avg_hours_per_delay=self.input_dict['avg_hours_per_delay'],
                                              std_dev_hours_per_delay=self.input_dict['std_dev_hours_per_delay'],
                                              delay_speed_m_per_s=self.input_dict['delay_speed_m_per_s'],
                                              seed=self.input_dict['seed'])

        self.input_dict['start_delay_hours'] = 100
        self.input_dict['mission_time_hours'] = 50
        self.input_dict['critical_wind_speed_m_per_s'] = 8.0
        self.input_dict['wind_height_of_interest_m'] = 100
        self.input_dict['wind_shear_exponent'] = 0.25
        self.input_dict['season_construct'] = ['spring', 'summer']
        self.input_dict['time_construct'] = 'normal'
        self.input_dict['operational_hrs_per_day'] = 10
        self.input_dict['critical_speed_non_erection_wind_delays_m_per_s'] = 10
        self.input_dict['critical_height_non_erection_wind_delays_m'] = 10

        #Calculate Costs:
        material_price = os.path.join(landbosse_test_input_dir(), 'material_price_proprietary.csv')
        self.input_dict['material_price'] = pd.read_csv(material_price)

        rsmeans_csv = os.path.join(landbosse_test_input_dir(), 'rsmeans_proprietary.csv')
        self.input_dict['rsmeans'] = pd.read_csv(rsmeans_csv)
        self.input_dict['rsmeans_per_diem'] = 144



        self.input_dict['overtime_multiplier'] = 1.4


        self.output_dict = dict()
        self.foundation_cost = FoundationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)



    def test_FoundationCostModule(self):
        """
        Black box test to check whether module is ran successfully or not
        """
        run_FoundationCost = FoundationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        trial_run = run_FoundationCost.run_module()


        if trial_run[0] == 0 :
            print('\n\n================== MODULE EXECUTION SUCCESS =========================\n')
            print(' FoundationCost module ran successfully. See the list of inputs'
                  '\n and outputs below used by the module in its calculations:')
            print( '\n=====================================================================\n')

        elif trial_run[0] == 1 :
            print('\n\n====================================== MODULE EXECUTION FAILURE ======================================\n')
            print(' > FoundationCost module failed to run successfully. Error detected: ',trial_run[1] ,
                  '\n > Scroll below to see detailed information about error encountered.'
                  '\n > See the list of inputs below used by the module in its calculations:')
            print('\n========================================================================================================\n')


        print('\nGiven below is the set of inputs fed into FoundationCost module:\n')
        for key, value in self.input_dict.items():
            if isinstance(value, pd.DataFrame):
                print(key, ':\n', value)
            else:
                print(key, ':', value)

        if trial_run[0] == 0:  # Only print outputs if module ran successfully.
            print('\nGiven below is the set of outputs calculated by the FoundationCost module:\n')
            for key, value in self.output_dict.items():
                if isinstance(value, pd.DataFrame):
                    print('\nNow printing DataFrame ->', key, ':\n', value)
                else:
                    print(key, ':', value)



import math
import traceback
import pytest
import traceback

class ManagementCost:
    """
    This class models management costs of a wind plant. Its inputs are
    configured with a dictionary with the key value pairs being the
    inputs into the model.

    See the input_output documentation below for a description of the key
    names for this inputs values dictionary.

    The INPUT keys are the following:

    project_value_usd
        (float) Sum of all other BOS costs (e.g., roads, foundations, erection)

    foundation_cost_usd
        (float) Foundation costs for the project

    construct_duration
        (float) Project duration (in months)

    num_hwy_permits
        (int) Number of highway permits needed for the project

    markup_constants
        (dict) Markup and contingency costs that can be set by user
        (see the markup_contingency method for key names to use in this dictionary)

    num_turbines
        (int) Number of turbines for project

    project_size_megawatts
        (float) Total power output of project in megawatts

    hub_height_meters
        (float) Hub height for all turbines in meters

    project_size_megawatts

        (float) Total power output of project in megawatts

    num_access_roads
        (int) Number of access roads into project site

    site_facility_building_area_df
        (pd.DataFrame) Building areas dataframe. This should be loaded
        from a .csv file on the filesystem.

    The OUTPUT keys are the following

    insurance
        (float) The cost of insurance for the project. USD. Eqn. 3.4.1

    construction_permitting
        (float) The cost of the construction permitting. USD.

    project_management
        (float) The cost of the project management. USD.

    bonding
        (float) The cost of binding for the project. USD.

    engineering_usd
        (float) Site-specific engineering costs for foundations and collection. USD.

    site_security_usd
        (float) estimate cost of site security. USD.

    site_facility
        (float) Site facility costs. USD.

    markup_contingency
        (float) Markup contingency costs. USD.

    total_cost
        (float) Total cost of everything else
    """

    def __init__(self, input_dict, output_dict, project_name):
        """
        This method runs all cost calculations in the model based on the
        inputs key value pairs in the dictionary that is the only
        argument.

        Parameters
        ----------
        input_dict : dict
            Dictionary with the inputs key / value pairs

        output_dict : dict
            Dictionary with output key / value pairs.
        """
        self.in_distributed_mode = 'override_total_management_cost' in input_dict
        self.validate_inputs(input_dict)
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

    def validate_inputs(self, input_dict):
        """
        This method checks a dictionary to make sure it has keys for all
        necessary values needed for calculations in an instance of this
        class. It is made to validate input_dict dictionaries.

        Returns
        -------
        None
            This method returns nothing if the validation passes. If it does
            not pass an exception is raised.

        Raises
        ------
        ValueError
            If one of the keys is missing, this method raises a ValueError
        """
        if not self.in_distributed_mode:
            required_keys = {
                'project_value_usd',
                'foundation_cost_usd',
                'construct_duration',
                'num_hwy_permits',
                'num_turbines',
                'project_size_megawatts',
                'hub_height_meters',
                'num_access_roads',
                'markup_contingency',
                'markup_warranty_management',
                'markup_sales_and_use_tax',
                'markup_overhead',
                'markup_profit_margin',
                'site_facility_building_area_df'
            }
            found_keys = set(input_dict.keys())
            if len(required_keys - found_keys) > 0:
                err_msg = '{}: did not find all required keys in inputs dictionary. Missing keys are {}'
                raise ValueError(err_msg.format(type(self).__name__, required_keys - found_keys))

    def insurance(self):
        """
        Calculate insurance costs based on project value, builder size, and project size;
        equation based on empirical data from industry.

        Eqn. 3.4.1

        Includes:

        Builder's risk

        General liability

        Umbrella policy

        Professional liability

        It uses only the self.project_value attribute in calculations.

        :math:`C_I = 0.0056 * V_p`

        Returns
        -------
        float
            Insurance costs in USD
        """
        insurance_cost = 0.0056 * self.input_dict['project_value_usd']
        return insurance_cost

    def construction_permitting(self):
        """
        Calculate construction permitting costs based empirical data from industry.
        Includes building and highway permits

        Returns
        -------
        float
            Construction permitting cost in USD
        """
        building_permits = 0.02 * self.input_dict['foundation_cost_usd']
        highway_permits = 20000 * self.input_dict['num_hwy_permits']
        construction_permitting_cost = building_permits + highway_permits
        return construction_permitting_cost

    def bonding(self):
        """
        Calculate bonding costs based on project size; equation based on empirical data from industry.

        Returns
        -------
        float
            Bonding cost in USD
        """
        # Calculate bonding costs based on project size
        performance_bond_cost = 0.01 * self.input_dict['project_value_usd']
        return performance_bond_cost

    def project_management(self):
        """
        Calculate project management costs based on project size; equation based on empirical data from industry.
        Includes:

        Project manager and assistant project manager for site

        Site managers for civil, electrical, and erection

        QA/QC management

        QA/QC inspections for civil, structural, electrical, and mechanical

        Administrative support for the site

        Health and safety supervisors

        Environmental supervisors

        Office equipment & materials

        Site radios, communication, and vehicles

        Management team per diem and travel

        Legal and public relations

        Returns
        -------
        float
            Project management cost
        """
        # todo: add relationship to site-specific interface with public infrastructure
        if self.output_dict['actual_construction_months'] < 28:
            project_management_cost = (53.333 * self.output_dict['actual_construction_months'] ** 2 -
                                       3442 * self.output_dict['actual_construction_months'] +
                                       209542) * (self.output_dict['actual_construction_months'] + 2)
        else:
            project_management_cost = (self.output_dict['actual_construction_months'] + 2) * 155000
        return project_management_cost

    def markup_contingency(self):
        """
        Calculate mark-up and contingency costs based on project value. Includes:

        Markup contingency

        Markup warranty management

        Sales and use tax

        Markup overhead

        Markup profit margin

        Returns
        -------
        float
            Mark up and contingency costs.
        """
        # Calculate mark-up and contingency costs based on project value
        markup_contingency_cost = (self.input_dict['markup_contingency']
                                   + self.input_dict['markup_warranty_management']
                                   + self.input_dict['markup_sales_and_use_tax']
                                   + self.input_dict['markup_overhead']
                                   + self.input_dict['markup_profit_margin']) * self.input_dict['project_value_usd']
        return markup_contingency_cost

    def engineering_foundations_collection_sys(self):
        """
        Calculate site-specific engineering costs for foundations and collection system
        based on empirical data from industry. Includes met masts and power performance.

        Returns
        -------
        float
            site-specific engineering costs
        """
        # development engineering costs for foundations and collection system
        if self.input_dict['project_size_megawatts'] < 200:
            development_engineering_cost = 7188.5 * self.input_dict['num_turbines'] + round(3.4893 * math.log(self.input_dict['num_turbines']) - 7.3049, 0) * 16800 + 165675
        else:
            development_engineering_cost = 7188.5 * self.input_dict['num_turbines'] + round(3.4893 * math.log(self.input_dict['num_turbines']) - 7.3049, 0) * 16800 + 327250

        # engineering costs for met masts
        # TODO: Projects less than 30 MW for met masts
        if 30 <= self.input_dict['project_size_megawatts'] <= 100:
            num_perm_met_mast = 2
            num_temp_met_mast = 2
        elif 100 < self.input_dict['project_size_megawatts'] <= 300:
            num_perm_met_mast = 2
            num_temp_met_mast = 4
        elif self.input_dict['project_size_megawatts'] > 300:
            num_perm_met_mast = round(self.input_dict['project_size_megawatts'] / 100)
            num_temp_met_mast = round(self.input_dict['project_size_megawatts'] / 100) * 2
        else:
            num_perm_met_mast = 1
            num_temp_met_mast = 1

        if self.input_dict['hub_height_meters'] < 90:
            multiplier_perm = 232600
            multiplier_temp = 92600
        else:
            multiplier_perm = 290000
            multiplier_temp = 116800

        met_mast_cost = (num_perm_met_mast * multiplier_perm) + (num_temp_met_mast * multiplier_temp) + 200000

        engineering_cost = development_engineering_cost + met_mast_cost

        return engineering_cost

    def site_facility(self):
        """
        Uses empirical data to estimate cost of site facilities and security, including


        Site facilities:

        Building design and construction

        Drilling and installing a water well, including piping

        Electric power for a water well

        Septic tank and drain field


        Site security:

        Constructing and reinstating the compound

        Constructing and reinstating the batch plant site

        Setting up and removing the site offices for the contractor, turbine supplier, and owner

        Restroom facilities

        Electrical and telephone hook-up

        Monthly office costs

        Signage for project information, safety and directions

        Cattle guards and gates

        Number of access roads



        In main.py, a csv is loaded into a Pandas dataframe. The columns of the
        dataframe must be:

        Size Min (MW)
            Minimum power output for a plant that needs a certain size of
            building.

        Size Max (MW)
            Maximum power output of a plant that need a certain size of
            building.

        Building Area (sq. ft.)
            The area of the building needed to provide O & M support to plants
            with power output between "Size Min (MW)" and "Size Max (MW)".

        Returns
        -------
        float
            Building area in square feet
        """
        df = self.input_dict['site_facility_building_area_df']
        project_size_megawatts = self.input_dict['project_size_megawatts']
        row = df[(df['Size Max (MW)'] > project_size_megawatts) & (df['Size Min (MW)'] <= project_size_megawatts)]
        building_area_sq_ft = float(row['Building area (sq. ft.)'])

        construction_building_cost = building_area_sq_ft * 125 + 176125

        ps = self.input_dict['project_size_megawatts']
        ct = self.output_dict['actual_construction_months']
        nt = self.input_dict['num_turbines']
        if nt < 30:
            nr = 1
            acs = 30000
        elif nt < 100:
            nr = round(0.05 * nt)
            acs = 240000
        else:
            nr = round(0.05 * nt)
            acs = 390000
        compound_security_cost = 9825 * nr + 29850 * ct + acs + 60 * ps + 62400

        site_facility_cost = construction_building_cost + compound_security_cost
        return site_facility_cost


    def total_management_cost(self):
        """
        Calculates the total cost of returned by the rest of the methods.
        This must be called after all the other methods in the module.

        This method uses the outputs for all the other individual costs as
        set in self.output_dict. It does not call the other methods directly.

        Returns
        -------
        float
            Total management cost as summed from the outputs of all
            other methods.
        """
        total = 0
        total += self.output_dict['insurance_usd']
        total += self.output_dict['construction_permitting_usd']
        total += self.output_dict['bonding_usd']
        total += self.output_dict['project_management_usd']
        total += self.output_dict['markup_contingency_usd']
        total += self.output_dict['engineering_usd']
        total += self.output_dict['site_facility_usd']
        return total

    def outputs_for_detailed_tab(self):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """
        result = []
        if self.in_distributed_mode:
            row = {
                'project_id_with_serial': self.project_name,
                'module': type(self).__name__,
                'type': 'variable',
                'variable_df_key_col_name': 'total_management_cost',
                'unit': 'usd',
                'value': self.output_dict['total_management_cost']
            }
            result.append(row)
        else:
            management_cost_keys = [
                'insurance_usd',
                'construction_permitting_usd',
                'bonding_usd',
                'project_management_usd',
                'markup_contingency_usd',
                'engineering_usd',
                'site_facility_usd'
            ]

            for key in management_cost_keys:
                value = self.output_dict[key]
                row = {
                    'project_id_with_serial': self.project_name,
                    'module': type(self).__name__,
                    'type': 'variable',
                    'variable_df_key_col_name': key,
                    'unit': 'usd',
                    'value': value
                }
                result.append(row)

        return result

    def outputs_for_module_type_operation(self):
        """
        Outputs dictionaries that are rows for the
        costs_by_module_type_operation

        Returns
        -------
        list
            List of dicts, with each dict representing a row for
            the output.
        """
        result = []
        module = type(self).__name__
        turbine_rating_MW = self.input_dict['turbine_rating_MW']
        num_turbines = self.input_dict['num_turbines']
        project_size_kw = num_turbines * turbine_rating_MW * 1000

        if self.in_distributed_mode:
            result.append({
                'type_of_cost': 'total_management_cost',
                'raw_cost': self.output_dict['total_management_cost']
            })

        else:
            result.append({
                'type_of_cost': 'insurance',
                'raw_cost': self.output_dict['insurance_usd']
            })
            result.append({
                'type_of_cost': 'Construction Permitting',
                'raw_cost': self.output_dict['construction_permitting_usd']
            })
            result.append({
                'type_of_cost': 'Project Management',
                'raw_cost': self.output_dict['project_management_usd']
            })
            result.append({
                'type_of_cost': 'Bonding',
                'raw_cost': self.output_dict['bonding_usd']
            })
            result.append({
                'type_of_cost': 'Markup Contingency',
                'raw_cost': self.output_dict['markup_contingency_usd']
            })
            result.append({
                'type_of_cost': 'Engineering Foundation and Collections System (includes met mast)',
                'raw_cost': self.output_dict['engineering_usd']
            })
            result.append({
                'type_of_cost': 'Site Facility',
                'raw_cost': self.output_dict['site_facility_usd']
            })

        for _dict in result:
            _dict['turbine_rating_MW'] = self.input_dict['turbine_rating_MW']
            _dict['num_turbines'] = self.input_dict['num_turbines']
            _dict['rotor_diameter_m'] = self.input_dict['rotor_diameter_m']
            _dict['project_id_with_serial'] = self.project_name
            _dict['operation_id'] = 'Management'
            _dict['module'] = module
            _dict['raw_cost_total_or_per_turbine'] = 'total'
            _dict['cost_per_turbine'] = _dict['raw_cost'] / num_turbines
            _dict['cost_per_project'] = _dict['raw_cost']
            _dict['usd_per_kw_per_project'] = _dict['raw_cost'] / project_size_kw

        return result

    def run_module(self):
        """
        Runs all the calculation methods in order.

        Parameters
        ----------
        project_name : str
            The name of the project for which this calculation is being done.

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            depends on the error condition. If the condition is 1, then the second
            element is the error raised that caused the failure. If the condition is
            0 then the second element is 0 as well.
        """
        try:
            if self.in_distributed_mode:
                self.output_dict['insurance_usd'] = 0
                self.output_dict['construction_permitting_usd'] = 0
                self.output_dict['project_management_usd'] = 0
                self.output_dict['bonding_usd'] = 0
                self.output_dict['markup_contingency_usd'] = 0
                self.output_dict['engineering_usd'] = 0
                self.output_dict['site_facility_usd'] = 0
                self.output_dict['total_management_cost'] = self.input_dict['override_total_management_cost']

            else:
                self.output_dict['insurance_usd'] = self.insurance()
                self.output_dict['construction_permitting_usd'] = self.construction_permitting()
                self.output_dict['project_management_usd'] = self.project_management()
                self.output_dict['bonding_usd'] = self.bonding()
                self.output_dict['markup_contingency_usd'] = self.markup_contingency()
                self.output_dict['engineering_usd'] = self.engineering_foundations_collection_sys()
                self.output_dict['site_facility_usd'] = self.site_facility()
                self.output_dict['total_management_cost'] = self.total_management_cost()
            self.output_dict['management_cost_csv'] = self.outputs_for_detailed_tab()
            self.output_dict['mangement_module_type_operation'] = self.outputs_for_module_type_operation()

            return 0, 0    # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} ManagementCost")
            self.input_dict['error']['ManagementCost'] = error
            return 1, error  # module did not run successfully



import traceback
import math

from .ManagementCost import ManagementCost
from .FoundationCost import FoundationCost
from .SubstationCost import SubstationCost
from .GridConnectionCost import GridConnectionCost
from .SitePreparationCost import SitePreparationCost
from .CollectionCost import Cable, Array, ArraySystem
from .ErectionCost import ErectionCost
from .DevelopmentCost import DevelopmentCost


class Manager:
    """
    The Manager class distributes input and output dictionaries among
    the various modules. It maintains the hierarchical dictionary
    structure.
    """

    def __init__(self, input_dict, output_dict):
        """
        This initializer sets up the instance variables of:

        self.cost_modules: A list of cost module instances. Each of the
            instances must implement the method input_output.

        self.input_dict: A placeholder for the inputs dictionary

        self.output_dict: A placeholder for the output dictionary
        """
        self.input_dict = input_dict
        self.output_dict = output_dict

    def execute_landbosse(self, project_name):
        try:
            # Create weather window that will be used for all tasks (window for entire project; selected to restrict to seasons and hours specified)
            weather_data_user_input = self.input_dict['weather_window']
            season_construct = self.input_dict['season_construct']
            time_construct = self.input_dict['time_construct']
            daily_operational_hours = self.input_dict['hour_day'][time_construct]

            # Filtered window. Restrict to the seasons and hours specified.
            filtered_weather_window = weather_data_user_input.loc[(weather_data_user_input['Season'].isin(season_construct)) & (weather_data_user_input['Time window'] == time_construct)]
            filtered_weather_window = filtered_weather_window[0:(math.ceil(self.input_dict['construct_duration'] * 30 * daily_operational_hours))]

            # Rename weather data to specify types
            self.input_dict['weather_window'] = filtered_weather_window
            self.input_dict['weather_data_user_input'] = weather_data_user_input

            foundation_cost = FoundationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=project_name)
            foundation_cost.run_module()

            roads_cost = SitePreparationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=project_name)
            roads_cost.run_module()

            substation_cost = SubstationCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=project_name)
            substation_cost.run_module()

            transdist_cost = GridConnectionCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=project_name)
            transdist_cost.run_module()

            collection_cost = ArraySystem(input_dict=self.input_dict, output_dict=self.output_dict, project_name=project_name)
            collection_cost.run_module()

            development_cost = DevelopmentCost(input_dict=self.input_dict, output_dict=self.output_dict,
                                          project_name=project_name)
            development_cost.run_module()

            erection_cost_output_dict = dict()
            erection_cost = ErectionCost(
                input_dict=self.input_dict,
                output_dict=self.output_dict,
                project_name=project_name
            )
            erection_cost.run_module()

            self.output_dict['erection_cost'] = erection_cost_output_dict



            self.output_dict['actual_construction_months'] = self.output_dict['siteprep_construction_months'] + \
                                                             max(self.output_dict['erection_construction_months'],
                                                             self.output_dict['foundation_construction_months'],
                                                             self.output_dict['collection_construction_months']) + 1

            if self.output_dict['actual_construction_months'] < self.input_dict['construct_duration']:
                road_cost = self.output_dict['total_road_cost']
                index = road_cost['Type of cost'] == 'Other'
                other = road_cost[index]
                amount_shorter_than_input_construction_time = (self.input_dict['construct_duration'] - self.output_dict['siteprep_construction_months'])
                road_cost.at[index, 'Cost USD'] = other['Cost USD'] - amount_shorter_than_input_construction_time * 55500
                self.output_dict['total_road_cost'] = road_cost

            total_costs = self.output_dict['total_collection_cost']
            total_costs = total_costs.append(self.output_dict['total_road_cost'], sort=False)
            total_costs = total_costs.append(self.output_dict['total_transdist_cost'], sort=False)
            total_costs = total_costs.append(self.output_dict['total_substation_cost'], sort=False)
            total_costs = total_costs.append(self.output_dict['total_foundation_cost'], sort=False)
            total_costs = total_costs.append(self.output_dict['total_erection_cost'], sort=False)
            total_costs = total_costs.append(self.output_dict['total_development_cost'], sort=False)



            self.output_dict['erection_cost'] = erection_cost_output_dict

            try:
                total_costs = self.output_dict['total_collection_cost']
                total_costs = total_costs.append(self.output_dict['total_road_cost'], sort=False)
                total_costs = total_costs.append(self.output_dict['total_transdist_cost'], sort=False)
                total_costs = total_costs.append(self.output_dict['total_substation_cost'], sort=False)
                total_costs = total_costs.append(self.output_dict['total_foundation_cost'], sort=False)
                total_costs = total_costs.append(self.output_dict['total_erection_cost'], sort=False)
                total_costs = total_costs.append(self.output_dict['total_development_cost'], sort=False)

                self.input_dict['project_value_usd'] = total_costs.sum(numeric_only=True)[0]
                self.output_dict['project_value_usd'] = self.input_dict['project_value_usd']
                self.input_dict['foundation_cost_usd'] = self.output_dict['total_foundation_cost'].sum(numeric_only=True)[0]

                management_cost = ManagementCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=project_name)
                management_cost.run_module()
                foo = total_costs.sum(numeric_only=True)[0]
                self.input_dict['project_value_usd'] = total_costs.sum(numeric_only=True)[0] + \
                                                       self.output_dict['total_management_cost']

                self.output_dict['project_value_usd'] = self.input_dict['project_value_usd']

            except Exception as error:  # exception handling for landbosse_api
                self.input_dict['error']['total_cost_error'] = error
            return 0

        except Exception as error:
            traceback.print_exc()
            self.input_dict['error']['Manager'] = error
            return 1  # module did not run successfully



import traceback
import pandas as pd
import math


from .CostModule import CostModule


class GridConnectionCost(CostModule):
    """
    TransDistCost.py
     - Created by Annika Eberle and Owen Roberts on Dec. 17, 2018
     - Refactored by Parangat Bhaskar and Alicia Key on June 3, 2019

    Calculates the costs associated with transmission and distribution for land-based wind projects (module is currently based on curve fit of empirical data)

    * Get distance to interconnection
    * Get interconnection voltage
    * Get toggle for new switchyard
    * Return total transmission and distribution costs

    \n\n**Keys in the input dictionary are the following:**

    distance_to_interconnect_mi
        (float) distance to interconnection [in miles]

    \n\n**Keys in the output dictionary are the following:**

    trans_dist_usd
        (float) total transmission and distribution costs [in USD]

    """

    def __init__(self, input_dict, output_dict , project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

    def calculate_costs(self, calculate_costs_input_dict, calculate_costs_output_dict):
        """
        Function to calculate total costs for transmission and distribution.

        Parameters
        ----------
        <None>

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            either returns a 0 if the module ran successfully, or it returns the error
            raised that caused the failure.
        """
        if 'grid_system_size_MW' not in calculate_costs_input_dict:
            project_size = calculate_costs_input_dict['turbine_rating_MW'] * \
                           calculate_costs_input_dict['num_turbines']
        else:
            project_size = calculate_costs_input_dict['grid_system_size_MW']

        # Switch between utility scale model and distributed model
        # Run utility version of GridConnectionCost for project size > 15 MW:
        if project_size > 15:
            if calculate_costs_input_dict['distance_to_interconnect_mi'] == 0:
                calculate_costs_output_dict['trans_dist_usd'] = 0
            else:
                if calculate_costs_input_dict['new_switchyard'] is True:
                    calculate_costs_output_dict['interconnect_adder_USD'] = 18115 * self.input_dict['interconnect_voltage_kV'] + 165944
                else:
                    calculate_costs_output_dict['interconnect_adder_USD'] = 0
                calculate_costs_output_dict['trans_dist_usd'] = ((1176 * self.input_dict[
                    'interconnect_voltage_kV'] + 218257) * (calculate_costs_input_dict['distance_to_interconnect_mi'] ** (
                    -0.1063)) * calculate_costs_input_dict['distance_to_interconnect_mi']) + calculate_costs_output_dict[
                                                                    'interconnect_adder_USD']

        # Run distributed wind version of GridConnectionCost for project size < 15 MW:
        else:
            # Code below is for newer version of LandBOSSE which incorporates distributed wind into the model:
            calculate_costs_output_dict['tower_to_point_of_interconnection_usd_per_kw'] = \
                1736.7 * ((project_size * 1000) ** (-0.272))

            calculate_costs_output_dict['trans_dist_usd'] = project_size * 1000 * \
                                                            calculate_costs_output_dict[
                                                                'tower_to_point_of_interconnection_usd_per_kw']


        calculate_costs_output_dict['trans_dist_usd_df'] = pd.DataFrame([['Other', calculate_costs_output_dict['trans_dist_usd'], 'Transmission and Distribution']],
                     columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        calculate_costs_output_dict['total_transdist_cost'] = calculate_costs_output_dict['trans_dist_usd_df']

        return calculate_costs_output_dict['trans_dist_usd_df']


    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """
        result = []
        module = type(self).__name__

        for row in self.output_dict['trans_dist_usd_df'].itertuples():
            dashed_row = '{} <--> {} <--> {}'.format(row[1], row[3], math.ceil(row[2]))
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'Type of Cost <--> Phase of Construction <--> Cost in USD ',
                'value': dashed_row,
                'last_number': row[2]
            })

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module

        self.output_dict['trans_dist_cost_csv'] = result
        return result

    def outputs_for_module_type_operation(self, input_dict, output_dict):
        """
        Outputs dictionaries that are rows for the
        costs_by_module_type_operation

        Returns
        -------
        list
            List of dicts, with each dict representing a row for
            the output.
        """
        result = []
        module = type(self).__name__

        costs_by_module_type_operation = output_dict['trans_dist_usd_df']
        for _, row in costs_by_module_type_operation.iterrows():
            _dict = dict()
            row = row.to_dict()
            _dict['operation_id'] = row['Phase of construction']
            _dict['type_of_cost'] = row['Type of cost']
            _dict['cost'] = row['Cost USD']
            result.append(_dict)

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module
            _dict['total_or_turbine'] = 'total'

        output_dict['trans_dist_cost_module_type_operation'] = result
        return result

    def run_module(self):
        """
        Runs the TransDistCost module and populates the IO dictionaries with calculated values.

        Parameters
        ----------
        <None>

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            either returns a 0 if the module ran successfully, or it returns the error
            raised that caused the failure.

        """
        try:
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            self.output_dict['trans_dist_cost_module_type_operation'] = \
                self.outputs_for_costs_by_module_type_operation(input_df=self.output_dict['trans_dist_usd_df'],
                                                                project_id=self.project_name,
                                                                total_or_turbine=True)
            return 0, 0 # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} GridConnectionCost")
            self.input_dict['error']['GridConnectionCost'] = error
            return 1, error # module did not run successfully

